import sys
from dataclasses import dataclass
from datetime import datetime, time
from decimal import Decimal
from enum import Enum
from typing import Annotated, Any, Literal, Optional, Union
from unittest import mock
from uuid import UUID

import pytest
from serpyco_rs import Serializer
from serpyco_rs.metadata import Alias, CamelCase, Discriminator, Max, MaxLength, Min, MinLength, OmitNone


def test_to_json_schema():
    class EnumCls(Enum):
        a = "a"

    @dataclass
    class InnerData:
        """Some important entity"""

        foo_filed: str

    @dataclass
    class OtherInnerData:
        """OtherInnerData"""

        optional_filed: Optional[InnerData] = None

    @dataclass
    class Data:
        """Docs"""

        a: Annotated[int, Min(0), Max(10)]
        """A field with bounds"""
        b: float
        c: Decimal
        d: bool
        e: Annotated[str, MinLength(1), MaxLength(5)]
        f: UUID
        g: time
        h: datetime
        i: EnumCls
        j: InnerData
        k: list[int]
        l: tuple[int, str, InnerData]
        m: dict[str, int]
        n: Any
        o: Annotated[InnerData, CamelCase]
        p: Annotated[OtherInnerData, OmitNone]
        some_filed: Annotated[str, Alias("fooFiled")]

    serializer = Serializer(Data)

    assert serializer.get_json_schema() == {
        "$ref": "#/components/schemas/tests.json_schema.test_convert.Data[no_format,keep_nones]",
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "components": {
            "schemas": {
                "tests.json_schema.test_convert.Data[no_format,keep_nones]": {
                    "description": "Docs",
                    "properties": {
                        "a": {
                            "description": "A " "field " "with " "bounds",
                            "maximum": 10,
                            "minimum": 0,
                            "type": "integer",
                        },
                        "b": {"type": "number"},
                        "c": {"oneOf": [{"type": "string"}, {"type": "number"}]},
                        "d": {"type": "boolean"},
                        "e": {"maxLength": 5, "minLength": 1, "type": "string"},
                        "f": {"format": "uuid", "type": "string"},
                        "g": {
                            "pattern": "^[0-9][0-9]:[0-9][0-9](:[0-9][0-9](\\.[0-9]+)?)??(([+-][0-9][0-9]:?[0-9][0-9])|Z)?$",
                            "type": "string",
                        },
                        "h": {
                            "pattern": "^[0-9]{4}-[0-9][0-9]-[0-9][0-9]T[0-9][0-9]:[0-9][0-9]:[0-9][0-9](\\.[0-9]+)?(([+-][0-9][0-9]:[0-9][0-9])|Z)?$",
                            "type": "string",
                        },
                        "i": {"enum": ["a"]},
                        "j": {
                            "$ref": "#/components/schemas/tests.json_schema.test_convert.InnerData[no_format,keep_nones]"
                        },
                        "k": {"items": {"type": "integer"}, "type": "array"},
                        "l": {
                            "maxItems": 3,
                            "minItems": 3,
                            "prefixItems": [
                                {"type": "integer"},
                                {"type": "string"},
                                {
                                    "$ref": "#/components/schemas/tests.json_schema.test_convert.InnerData[no_format,keep_nones]"
                                },
                            ],
                            "type": "array",
                        },
                        "m": {"additionalProperties": {"type": "integer"}, "type": "object"},
                        "n": {},
                        "p": {
                            "$ref": "#/components/schemas/tests.json_schema.test_convert.OtherInnerData[no_format,omit_nones]"
                        },
                        "o": {
                            "$ref": "#/components/schemas/tests.json_schema.test_convert.InnerData[camel_case,keep_nones]"
                        },
                        "fooFiled": {"type": "string"},
                    },
                    "required": [
                        "a",
                        "b",
                        "c",
                        "d",
                        "e",
                        "f",
                        "g",
                        "h",
                        "i",
                        "j",
                        "k",
                        "l",
                        "m",
                        "n",
                        "o",
                        "p",
                        "fooFiled",
                    ],
                    "type": "object",
                },
                "tests.json_schema.test_convert.InnerData[no_format,keep_nones]": {
                    "description": "Some " "important " "entity",
                    "properties": {"foo_filed": {"type": "string"}},
                    "required": ["foo_filed"],
                    "type": "object",
                },
                "tests.json_schema.test_convert.InnerData[no_format,omit_nones]": {
                    "description": "Some " "important " "entity",
                    "properties": {"foo_filed": {"type": "string"}},
                    "required": ["foo_filed"],
                    "type": "object",
                },
                "tests.json_schema.test_convert.InnerData[camel_case,keep_nones]": {
                    "description": "Some " "important " "entity",
                    "properties": {"fooFiled": {"type": "string"}},
                    "required": ["fooFiled"],
                    "type": "object",
                },
                "tests.json_schema.test_convert.OtherInnerData[no_format,omit_nones]": {
                    "description": "OtherInnerData",
                    "properties": {
                        "optional_filed": {
                            "anyOf": [
                                {"type": "null"},
                                {
                                    "$ref": "#/components/schemas/tests.json_schema.test_convert.InnerData[no_format,omit_nones]"
                                },
                            ]
                        }
                    },
                    "type": "object",
                },
            },
        },
    }


@pytest.mark.skipif(sys.version_info < (3, 10), reason="New style unions available after 3.10")
def test_to_json_schema__new_union_syntax():
    @dataclass
    class Data:
        """Docs"""

        a: int | None

    serializer = Serializer(Data)

    assert serializer.get_json_schema() == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$ref": "#/components/schemas/tests.json_schema.test_convert.Data[no_format,keep_nones]",
        "components": {
            "schemas": {
                "tests.json_schema.test_convert.Data[no_format,keep_nones]": {
                    "description": "Docs",
                    "properties": {"a": {"anyOf": [{"type": "null"}, {"type": "integer"}]}},
                    "required": ["a"],
                    "type": "object",
                }
            },
        },
    }


@dataclass
class TreeNode:
    """Node"""

    data: str
    left: Optional["TreeNode"] = None
    right: Optional["TreeNode"] = None


def test_to_json_schema__recursive_type():
    serializer = Serializer(TreeNode)

    assert serializer.get_json_schema() == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$ref": "#/components/schemas/tests.json_schema.test_convert.TreeNode[no_format,keep_nones]",
        "components": {
            "schemas": {
                "tests.json_schema.test_convert.TreeNode[no_format,keep_nones]": {
                    "description": "Node",
                    "properties": {
                        "data": {"type": "string"},
                        "left": {
                            "anyOf": [
                                {"type": "null"},
                                {
                                    "$ref": "#/components/schemas/tests.json_schema.test_convert.TreeNode[no_format,keep_nones]"
                                },
                            ]
                        },
                        "right": {
                            "anyOf": [
                                {"type": "null"},
                                {
                                    "$ref": "#/components/schemas/tests.json_schema.test_convert.TreeNode[no_format,keep_nones]"
                                },
                            ]
                        },
                    },
                    "required": ["data"],
                    "type": "object",
                }
            }
        },
    }


def test_to_json_schema__literal():
    serializer = Serializer(Literal["foo"])
    assert serializer.get_json_schema() == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "components": {"schemas": {}},
        "enum": ["foo"],
    }


def test_to_json_schema__tagged_union():
    @dataclass
    class Foo:
        val: int
        type: Literal["foo"]

    @dataclass
    class Bar:
        val: str
        type: Literal["bar"] = "bar"

    @dataclass
    class Base:
        field: Annotated[Union[Foo, Bar], Discriminator("type")]

    serializer = Serializer(Base)
    assert serializer.get_json_schema() == {
        "$ref": "#/components/schemas/tests.json_schema.test_convert.Base[no_format,keep_nones]",
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "components": {
            "schemas": {
                "tests.json_schema.test_convert.Bar[no_format,keep_nones]": {
                    "description": mock.ANY,
                    "properties": {"type": {"enum": ["bar"]}, "val": {"type": "string"}},
                    "required": ["val", "type"],
                    "type": "object",
                },
                "tests.json_schema.test_convert.Base[no_format,keep_nones]": {
                    "description": mock.ANY,
                    "properties": {
                        "field": {
                            "discriminator": {
                                "mapping": {
                                    "bar": "#/components/schemas/tests.json_schema.test_convert.Bar[no_format,keep_nones]",
                                    "foo": "#/components/schemas/tests.json_schema.test_convert.Foo[no_format,keep_nones]",
                                },
                                "propertyName": "type",
                            },
                            "oneOf": [
                                {
                                    "$ref": "#/components/schemas/tests.json_schema.test_convert.Foo[no_format,keep_nones]"
                                },
                                {
                                    "$ref": "#/components/schemas/tests.json_schema.test_convert.Bar[no_format,keep_nones]"
                                },
                            ],
                        }
                    },
                    "required": ["field"],
                    "type": "object",
                },
                "tests.json_schema.test_convert.Foo[no_format,keep_nones]": {
                    "description": mock.ANY,
                    "properties": {"type": {"enum": ["foo"]}, "val": {"type": "integer"}},
                    "required": ["val", "type"],
                    "type": "object",
                },
            }
        },
    }



from dataclasses import dataclass
from typing import Generic, Optional, TypeVar
from unittest import mock

from serpyco_rs import Serializer

T = TypeVar("T")


@dataclass
class GenericType(Generic[T]):
    value: Optional[T] = None
    path: Optional[str] = None


@dataclass
class CustomType:
    q: str
    w: int


@dataclass
class A:
    a: GenericType[bool]
    b: GenericType[CustomType]
    c: GenericType[int]
    d: GenericType[float]
    e: GenericType[str]
    f: GenericType[str]
    g: Optional[GenericType[str]] = None


def test_generics__serialization():
    serializer = Serializer(A, omit_none=True)

    a = A(
        a=GenericType(
            value=True,
            path="some_path",
        ),
        b=GenericType(
            value=CustomType(
                q="q",
                w=1,
            ),
            path="some_path",
        ),
        c=GenericType(
            value=1,
            path="some_path",
        ),
        d=GenericType(
            value=2.0,
            path="some_path",
        ),
        e=GenericType(
            value="value",
            path="some_path",
        ),
        f=GenericType(),
        g=None,
    )

    raw_a = {
        "a": {
            "value": True,
            "path": "some_path",
        },
        "b": {
            "value": {
                "q": "q",
                "w": 1,
            },
            "path": "some_path",
        },
        "c": {
            "value": 1,
            "path": "some_path",
        },
        "d": {
            "value": 2.0,
            "path": "some_path",
        },
        "e": {
            "value": "value",
            "path": "some_path",
        },
        "f": {},
    }

    assert serializer.dump(a) == raw_a
    assert serializer.load(raw_a) == a


def test_generics__swagger_schema():
    serializer = Serializer(A)
    assert serializer.get_json_schema() == {
        "$ref": "#/components/schemas/tests.test_generics.A[no_format,keep_nones]",
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "components": {
            "schemas": {
                "tests.test_generics.A[no_format,keep_nones]": {
                    "description": mock.ANY,
                    "properties": {
                        "a": {
                            "$ref": "#/components/schemas/tests.test_generics.tests.test_generics.GenericType[bool][no_format,keep_nones]"
                        },
                        "b": {
                            "$ref": "#/components/schemas/tests.test_generics.tests.test_generics.GenericType[tests.test_generics.CustomType][no_format,keep_nones]"
                        },
                        "c": {
                            "$ref": "#/components/schemas/tests.test_generics.tests.test_generics.GenericType[int][no_format,keep_nones]"
                        },
                        "d": {
                            "$ref": "#/components/schemas/tests.test_generics.tests.test_generics.GenericType[float][no_format,keep_nones]"
                        },
                        "e": {
                            "$ref": "#/components/schemas/tests.test_generics.tests.test_generics.GenericType[str][no_format,keep_nones]"
                        },
                        "f": {
                            "$ref": "#/components/schemas/tests.test_generics.tests.test_generics.GenericType[str][no_format,keep_nones]"
                        },
                        "g": {
                            "anyOf": [
                                {"type": "null"},
                                {
                                    "$ref": "#/components/schemas/tests.test_generics.tests.test_generics.GenericType[str][no_format,keep_nones]"
                                },
                            ]
                        },
                    },
                    "required": ["a", "b", "c", "d", "e", "f"],
                    "type": "object",
                },
                "tests.test_generics.CustomType[no_format,keep_nones]": {
                    "description": "CustomType(q: " "str, " "w: " "int)",
                    "properties": {"q": {"type": "string"}, "w": {"type": "integer"}},
                    "required": ["q", "w"],
                    "type": "object",
                },
                "tests.test_generics.tests.test_generics.GenericType[bool][no_format,keep_nones]": {
                    "description": mock.ANY,
                    "properties": {
                        "path": {"anyOf": [{"type": "null"}, {"type": "string"}]},
                        "value": {"anyOf": [{"type": "null"}, {"type": "boolean"}]},
                    },
                    "type": "object",
                },
                "tests.test_generics.tests.test_generics.GenericType[float][no_format,keep_nones]": {
                    "description": mock.ANY,
                    "properties": {
                        "path": {"anyOf": [{"type": "null"}, {"type": "string"}]},
                        "value": {"anyOf": [{"type": "null"}, {"type": "number"}]},
                    },
                    "type": "object",
                },
                "tests.test_generics.tests.test_generics.GenericType[int][no_format,keep_nones]": {
                    "description": mock.ANY,
                    "properties": {
                        "path": {"anyOf": [{"type": "null"}, {"type": "string"}]},
                        "value": {"anyOf": [{"type": "null"}, {"type": "integer"}]},
                    },
                    "type": "object",
                },
                "tests.test_generics.tests.test_generics.GenericType[str][no_format,keep_nones]": {
                    "description": mock.ANY,
                    "properties": {
                        "path": {"anyOf": [{"type": "null"}, {"type": "string"}]},
                        "value": {"anyOf": [{"type": "null"}, {"type": "string"}]},
                    },
                    "type": "object",
                },
                "tests.test_generics.tests.test_generics.GenericType[tests.test_generics.CustomType][no_format,keep_nones]": {
                    "description": mock.ANY,
                    "properties": {
                        "path": {"anyOf": [{"type": "null"}, {"type": "string"}]},
                        "value": {
                            "anyOf": [
                                {"type": "null"},
                                {"$ref": "#/components/schemas/tests.test_generics.CustomType[no_format,keep_nones]"},
                            ]
                        },
                    },
                    "type": "object",
                },
            }
        },
    }



import os

import nox

nox.options.sessions = ["test", "lint", "type_check", "rust_lint"]
nox.options.reuse_existing_virtualenvs = True


def build(session):
    if _is_ci():
        # Install form wheels
        session.install("--no-index", "--no-deps", "--find-links", "wheels/", "serpyco-rs")
        session.install("--find-links", "wheels/", "serpyco-rs")
        return

    session.run_always("maturin", "develop", "-r")


@nox.session
def test(session):
    build(session)
    session.install("-r", "requirements/dev.txt")
    session.run("pytest", "-vss", "tests/", *session.posargs)


@nox.session
def lint(session):
    build(session)
    session.install("-r", "requirements/lint.txt")

    session.cd("python/serpyco_rs")
    paths = [".", "../../tests", "../../bench"]
    session.run("black", *(["--check", "--diff", *paths] if _is_ci() else paths))
    session.run("isort", *(["--check", "--diff", *paths] if _is_ci() else paths))
    session.run("ruff", ".")


@nox.session(python=False)
def rust_lint(session):
    session.run("cargo", "fmt", "--all", *(["--", "--check"] if _is_ci() else []))
    session.run("cargo", "clippy", "--all-targets", "--all-features", "--", "-D", "warnings")


@nox.session
def type_check(session):
    build(session)
    session.install("-r", "requirements/type_check.txt")

    session.cd("python/serpyco_rs")
    session.run("pyright")
    session.run("pyright", "--verifytypes", "serpyco_rs")
    session.run("mypy", ".", "--strict", "--implicit-reexport", "--pretty")


@nox.session
def bench(session):
    build(session)
    session.install("-r", "requirements/bench.txt")

    session.run(
        "pytest",
        "--verbose",
        "--benchmark-min-time=0.25",
        "--benchmark-max-time=1",
        "--benchmark-disable-gc",
        "--benchmark-autosave",
        "--benchmark-save-data",
        "--benchmark-compare",
        "--ignore=bench/test_full.py",
        *(session.posargs if session.posargs else ["bench"]),
    )


@nox.session
def bench_codespeed(session):
    build(session)
    session.install("-r", "requirements/bench.txt")
    session.install('pytest-codspeed')
    session.run("pytest", "bench", "--ignore=bench/compare", "--codspeed")


def _is_ci() -> bool:
    return bool(os.environ.get("CI", None))
